# Copyright 2020 Novo Nordisk Foundation Center for Biosustainability,
# Technical University of Denmark.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Provide the main function to be used as an entry point for the CLI."""


import argparse
import logging
import sys
from typing import Optional

from . import __version__
from .api import get_latest_tag


logger = logging.getLogger("tag_spy")


def main() -> None:
    """Define the console script entry point and thus the command line interface."""
    parser = argparse.ArgumentParser(description="Find the latest Docker image tag.")
    parser.add_argument(
        "image",
        metavar="IMAGE",
        help="A Docker image specification, for example, dddecaf/wsgi-base.",
    )
    parser.add_argument(
        "tag",
        metavar="BASE_TAG",
        help="The first part of an image tag. So if your tags have the format "
        "'<image>:<base>_<date>_<commit>', you should supply the <base> here.",
    )
    parser.add_argument(
        "label",
        metavar="LABEL",
        help="The name of the label that contains the build timestamp, for example,"
        " 'com.business.build.timestamp'.",
    )
    parser.add_argument("--version", action="version", version=__version__)
    parser.add_argument(
        "--verbosity",
        help="The desired log level (default WARNING).",
        choices=("CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG"),
        default="WARNING",
    )
    parser.add_argument(
        "--authentication",
        metavar="URL",
        help="The URL from where to retrieve an access token for the registry "
        "(the default https://auth.docker.io/token corresponds to Docker Hub).",
        default="https://auth.docker.io/token",
    )
    parser.add_argument(
        "--registry",
        metavar="URL",
        help="The URL of the container registry API "
        "(default https://registry-1.docker.io).",
        default="https://registry-1.docker.io",
    )
    parser.add_argument(
        "--service",
        metavar="URI",
        help="The service URI for which to request an access token "
        "(default registry.docker.io).",
        default="registry.docker.io",
    )
    parser.add_argument(
        "--username",
        metavar="STR",
        help="A username used for authentication when retrieving an access token. "
        "Without a username no authentication will be attempted.",
        default=None,
    )
    parser.add_argument(
        "--password",
        action="store_true",
        help="A password will be requested interactively. It is used for "
        "authentication when retrieving an access token.The password field may "
        "itself be a token for Bearer authentication when the username is "
        "'oauth2accesstoken'.",
    )
    parser.add_argument(
        "--password-stdin",
        action="store_true",
        help="A password will be read from standard input. It is used for "
        "authentication when retrieving an access token.The password field may "
        "itself be a token for Bearer authentication when the username is "
        "'oauth2accesstoken'.",
    )
    args = parser.parse_args()
    logging.basicConfig(level=args.verbosity, format="[%(levelname)s] %(message)s")
    if args.password_stdin:
        password: Optional[str] = sys.stdin.read().rstrip("\r\n")
    else:
        password = input("Password:") if args.password else None
    print(
        get_latest_tag(
            image=args.image,
            base_tag=args.tag,
            label=args.label,
            authentication_url=args.authentication,
            registry_url=args.registry,
            service=args.service,
            username=args.username,
            password=password,
        )
    )



# Copyright 2020 Novo Nordisk Foundation Center for Biosustainability,
# Technical University of Denmark.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Provide helpers that interact with a Docker registry."""


import logging
from datetime import datetime, timezone
from operator import itemgetter
from typing import List
from urllib.parse import SplitResult, urlencode, urlunsplit
from urllib.request import OpenerDirector

from .http_helpers import get_response_json
from .image_tag_triple import ImageTagTriple


logger = logging.getLogger(__name__)


def get_token(
    opener: OpenerDirector, parts: SplitResult, image: str, service: str
) -> str:
    """
    Return an access token for the requested image and service.

    Args:
        opener (urllib.request.OpenerDirector): The opener with attached handlers and
            headers for calling the URL.
        parts (urllib.parse.SplitResult): The separate parts of the authentication URL
            as returned by ``urlsplit``.
        image (str): The fully specified image name, for example, 'dddecaf/wsgi-base'.
        service (str): The URL of the service for which to request an access token.

    Returns:
        str: The access token which is valid for a pre-specified amount of time.

    Raises:
        urllib.error.URLError: In case of problems communicating with the registry.

    """
    params = {"scope": f"repository:{image}:pull", "service": service}
    url = urlunsplit(parts._replace(query=urlencode(params)))
    logger.debug("Retrieving token at %r.", url)
    data = get_response_json(opener, url)
    return str(data["token"])


def verify_v2_capability(opener: OpenerDirector, parts: SplitResult) -> None:
    """
    Verify that the registry API actually supports version 2.

    Args:
        opener (urllib.request.OpenerDirector): The opener with attached handlers and
            headers for calling the URL.
        parts (urllib.parse.SplitResult): The separate parts of the registry API URL
            as returned by ``urlsplit``.

    Raises:
        urllib.error.URLError: In case the API does *not* support version 2.

    """
    url = urlunsplit(parts._replace(path="/v2/"))
    logger.debug("Verifying version 2 API capability at %r.", url)
    # The following statement will raise an URLError if the API does not support
    # version 2.
    # FIXME: Could be handled more nicely by raising an informative error.
    with opener.open(url) as response:
        assert response.status == 200


def get_tags(opener: OpenerDirector, parts: SplitResult, image: str) -> List[str]:
    """
    Return the list of tags for an image in its registry.

    Args:
        opener (urllib.request.OpenerDirector): The opener with attached handlers and
            headers for calling the URL.
        parts (urllib.parse.SplitResult): The separate parts of the registry API URL
            as returned by ``urlsplit``.
        image (str): The fully specified image name, for example, 'dddecaf/wsgi-base'.

    Returns:
        list: All the tags as strings that were found.

    Raises:
        urllib.error.URLError: In case of problems communicating with the registry.

    """
    url = urlunsplit(parts._replace(path=f"/v2/{image}/tags/list"))
    logger.debug("Retrieving image %r tags from %r.", image, url)
    logger.debug("%r", opener.addheaders)
    data = get_response_json(opener, url)
    return [str(t) for t in data["tags"]]


def get_image_digest(
    opener: OpenerDirector, parts: SplitResult, image: str, tag: str
) -> str:
    """
    Return an image's digest from its manifest.

    Args:
        opener (urllib.request.OpenerDirector): The opener with attached handlers and
            headers for calling the URL.
        parts (urllib.parse.SplitResult): The separate parts of the registry API URL
            as returned by ``urlsplit``.
        image (str): The fully specified image name, for example, 'dddecaf/wsgi-base'.
        tag (str): The base part of the tag that you are interested in, for
            example, 'alpine' will match 'dddecaf/wsgi-base:alpine_2020-04-28_24fe0a0'.

    Returns:
        str: The image's digest hash.

    Raises:
        urllib.error.URLError: In case of problems communicating with the registry.

    """
    url = urlunsplit(parts._replace(path=f"/v2/{image}/manifests/{tag}"))
    logger.info("Retrieving image %r digest from %r.", image, url)
    data = get_response_json(opener, url)
    return str(data["config"]["digest"])


def get_image_timestamp(
    opener: OpenerDirector,
    parts: SplitResult,
    image: str,
    digest: str,
    timestamp_label: str,
) -> datetime:
    """
    Return an image's build timestamp from its labels.

    Args:
        opener (urllib.request.OpenerDirector): The opener with attached handlers and
            headers for calling the URL.
        parts (urllib.parse.SplitResult): The separate parts of the registry API URL
            as returned by ``urlsplit``.
        image (str): The fully specified image name, for example, 'dddecaf/wsgi-base'.
        digest (str): An image digest as can be retrieved from its manifest.
        timestamp_label (str): The image label that defines the build timestamp,
            for example, 'dk.dtu.biosustain.wsgi-base.alpine.build.timestamp'.

    Returns:
        datetime.datetime: The timestamp that records when the image was built.

    Raises:
        urllib.error.URLError: In case of problems communicating with the registry.

    See Also:
        get_image_digest

    """
    url = urlunsplit(parts._replace(path=f"/v2/{image}/blobs/{digest}"))
    logger.info("Retrieving image %r configuration from %r.", image, url)
    data = get_response_json(opener, url)
    try:
        timestamp = datetime.fromisoformat(data["config"]["Labels"][timestamp_label])
    except KeyError:
        logger.error(
            "The requested label '%s' does not exist. Ignoring. "
            "Possibilities are: %s.",
            timestamp_label,
            ", ".join(data["config"]["Labels"].keys()),
        )
        timestamp = datetime.fromtimestamp(0, timezone.utc)
    return timestamp


def get_latest_by_timestamp(
    opener: OpenerDirector,
    parts: SplitResult,
    image: str,
    timestamp_label: str,
    tags: List[ImageTagTriple],
) -> ImageTagTriple:
    """
    Return the latest image tag as determined by the build timestamp.

    Args:
        opener (urllib.request.OpenerDirector): The opener with attached handlers and
            headers for calling the URL.
        parts (urllib.parse.SplitResult): The separate parts of the registry API URL
            as returned by ``urlsplit``.
        image (str): The fully specified image name, for example, 'dddecaf/wsgi-base'.
        timestamp_label (str): The image label that defines the build timestamp,
            for example, 'dk.dtu.biosustain.wsgi-base.alpine.build.timestamp'.
        tags (list): A collection of ``ImageTagTriple`` that all contain the same date.

    Returns:
        ImageTagTriple: The latest of the collection.

    Raises:
        urllib.error.URLError: In case of problems communicating with the registry.

    """
    latest = []
    for triple in tags:
        digest = get_image_digest(opener, parts, image, str(triple))
        logger.debug("Digest: %s", digest)
        build_timestamp = get_image_timestamp(
            opener, parts, image, digest, timestamp_label
        )
        logger.debug("%s: %s", timestamp_label, build_timestamp)
        latest.append((triple, build_timestamp))
    latest.sort(key=itemgetter(1), reverse=True)
    return latest[0][0]



# Copyright 2020 Novo Nordisk Foundation Center for Biosustainability,
# Technical University of Denmark.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Provide top level symbols."""


__author__ = (
    "Novo Nordisk Foundation Center for Biosustainability,"
    " Technical University of Denmark"
)
__email__ = "niso@biosustain.dtu.dk"
from tag_spy._version import get_versions

__version__ = get_versions()["version"]
del get_versions

from .api import get_latest_tag



# Copyright 2020 Novo Nordisk Foundation Center for Biosustainability,
# Technical University of Denmark.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Provide helpers that interact with a Docker registry."""

import gzip
import json
import logging
from typing import Optional, Union
from urllib.request import (
    HTTPBasicAuthHandler,
    HTTPPasswordMgrWithDefaultRealm,
    OpenerDirector,
    Request,
    build_opener,
)


logger = logging.getLogger(__name__)


def build_authentication_opener(
    registry_url: str, username: Optional[str] = None, password: Optional[str] = None,
) -> OpenerDirector:
    """
    Return an opener director that knows how to authenticate with the registry.

    Args:
        registry_url (str): The URL of the container registry API.
        username (str, optional): A username used for authentication when
            retrieving an access token. Without a username no authentication will be
            attempted.
        password (str, optional): A password used for authentication when
            retrieving an access token. The password field may itself be a token for
            Bearer authentication when the username is 'oauth2accesstoken'.

    Returns:
        OpenerDirector: A director for making HTTP requests. It's main method is `open`.

    """
    opener = build_opener()
    opener.addheaders.extend(
        [
            ("Accept", "application/vnd.docker.distribution.manifest.v2+json"),
            ("Accept-Encoding", "gzip, deflate"),
        ]
    )
    if username is not None:
        assert password is not None, "Authentication requires a password or token."
        # The standard library urllib.request has no HTTPBearerAuthHandler, thus we
        # create a corresponding header ourselves.
        if username.lower() == "oauth2accesstoken":
            opener.addheaders.append(("Authorization", f"Bearer {password}"))
        else:
            password_manager = HTTPPasswordMgrWithDefaultRealm()
            password_manager.add_password(
                None,  # type: ignore
                registry_url,
                username,
                password,
            )
            opener.add_handler(HTTPBasicAuthHandler(password_manager))
    return opener


def build_authenticated_opener(token: str,) -> OpenerDirector:
    """
    Return an opener director that authenticates requests with a Bearer token.

    Args:
        token (str): The access token.

    Returns:
        OpenerDirector: A director for making HTTP requests. It's main method is `open`.

    """
    opener = build_opener()
    opener.addheaders.extend(
        [
            ("Authorization", f"Bearer {token}"),
            ("Accept", "application/vnd.docker.distribution.manifest.v2+json"),
            ("Accept-Encoding", "gzip, deflate"),
        ]
    )
    return opener


def get_response_json(opener: OpenerDirector, url: Union[str, Request]) -> dict:
    """
    Perform an HTTP request and return the body deserialized as JSON.

    Args:
        opener (OpenerDirector): The director that executes the request.
        url (str or urllib.request.Request): The URL to open.

    Returns:
        dict: The JSON body as a dict object.

    Raises:
        urllib.error.HTTPError: In case there is a problem in HTTP communication.
        json.JSONDecodeError: In case the response body can not be deserialized.

    """
    with opener.open(url) as response:
        logger.debug("%d %s", response.status, response.reason)
        logger.debug("%r", response.getheaders())
        content = response.read()
        if response.getheader("Content-Encoding", "").lower() == "gzip":
            content = gzip.decompress(content)
    try:
        data = json.loads(content)
    except json.JSONDecodeError:
        logger.debug(content)
        raise
    return dict(data)



# Copyright 2020 Novo Nordisk Foundation Center for Biosustainability,
# Technical University of Denmark.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Provide a representation of a Docker image tag triple."""


from __future__ import annotations

from datetime import date
from typing import NamedTuple


class ImageTagTriple(NamedTuple):
    """Define a minimal class for storing complete tag information."""

    base: str
    date: date
    commit: str

    def __str__(self) -> str:
        """Return a string representation of the tag triple."""
        return f"{self.base}_{self.date.isoformat()}_{self.commit}"

    @classmethod
    def from_tag(cls, tag: str) -> ImageTagTriple:
        """
        Create a tag triple representation from a complete image tag.

        Args:
            tag (str): A Docker image tag of the form '<base>_<date>_<commit>'.

        Returns:
            ImageTagTriple: A new instance with separate tag parts.

        Raises:
            ValueError: In case the tag does not conform with the triple format or the
                <date> is not in iso-8601 format.

        """
        base, build_date, build_commit = tag.split("_")
        return cls(base=base, date=date.fromisoformat(build_date), commit=build_commit)



# Copyright 2020 Novo Nordisk Foundation Center for Biosustainability,
# Technical University of Denmark.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Define helpers that interact with a Docker registry."""


import logging
from itertools import takewhile
from operator import itemgetter
from typing import List

from .image_tag_triple import ImageTagTriple


logger = logging.getLogger(__name__)


def filter_latest_matching(all_tags: List[str], tag_part: str) -> List[ImageTagTriple]:
    """
    Return those tags that correspond to the expected format.

    Args:
        all_tags (list): All the tags as strings.
        tag_part (str): The base part of the tag that you are interested in, for
            example, 'alpine' will match 'dddecaf/wsgi-base:alpine_2020-04-28_24fe0a0'.

    Returns:
        list: A collection of ``ImageTagTriple`` instances.

    Raises:
        ValueError: In case no tags correspond to the expected format.

    """
    tags = []
    for tag in all_tags:
        try:
            triple = ImageTagTriple.from_tag(tag)
        except ValueError:
            continue
        if triple.base == tag_part:
            tags.append(triple)
    if len(tags) == 0:
        raise ValueError("No tags after filtering.")
    # Order tags with the latest date first.
    tags.sort(key=itemgetter(1), reverse=True)
    # Collect all tags created on the latest day.
    latest = tags[0].date
    return list(takewhile(lambda triple: triple.date == latest, tags))



# Copyright 2020 Novo Nordisk Foundation Center for Biosustainability,
# Technical University of Denmark.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Provide public API functions for handling Docker image tags."""


import logging
from typing import Optional
from urllib.parse import urlsplit

from .filter_helpers import filter_latest_matching
from .http_helpers import build_authenticated_opener, build_authentication_opener
from .registry_helpers import (
    get_latest_by_timestamp,
    get_tags,
    get_token,
    verify_v2_capability,
)


logger = logging.getLogger(__name__)


def get_latest_tag(
    image: str,
    base_tag: str,
    label: str,
    authentication_url: str = "https://auth.docker.io/token",
    registry_url: str = "https://registry-1.docker.io",
    service: str = "registry.docker.io",
    username: Optional[str] = None,
    password: Optional[str] = None,
) -> str:
    """
    Parse the latest DD-DeCaF-specific tag of a Docker image from its registry.

    Args:
        image (str): The fully specified image name, for example, 'dddecaf/wsgi-base'.
        base_tag (str): The base part of the tag that you are interested in, for
            example, 'alpine' will match 'dddecaf/wsgi-base:alpine_2020-04-28_24fe0a0'.
        label (str): The image label that defines the build timestamp, for example,
            'dk.dtu.biosustain.wsgi-base.alpine.build.timestamp'.
        authentication_url (str, optional): The URL from where to retrieve an access
            token for the registry (the default https://auth.docker.io/token
            corresponds to Docker Hub).
        registry_url (str, optional): The URL of the registry API
            (default https://registry-1.docker.io).
        service (str, optional): The service URI for which to request an access
            token (default registry.docker.io).
        username (str, optional): A username used for authentication when
            retrieving an access token. Without a username no authentication will be
            attempted.
        password (str, optional): A password used for authentication when
            retrieving an access token. The password field may itself be a token for
            Bearer authentication when the username is 'oauth2accesstoken'.

    Returns:
        str: The very latest Docker image tag in its entirety.

    Raises:
        RuntimeError: If there are unexpected complications with the tags of the
            specified image.
        urllib.error.URLError: In case of problems communicating with the registry.

    """
    auth_opener = build_authentication_opener(registry_url, username, password)
    token = get_token(auth_opener, urlsplit(authentication_url), image, service)
    opener = build_authenticated_opener(token)
    parts = urlsplit(registry_url)
    verify_v2_capability(opener, parts)
    tags = get_tags(opener, parts, image)
    if len(tags) == 0:
        raise RuntimeError(f"The requested image {image} does not have any tags.")
    try:
        latest_tags = filter_latest_matching(tags, base_tag)
    except ValueError:
        raise RuntimeError(
            f"The requested image {image} does not have any tags corresponding to the "
            f"expected format {base_tag}_<date>_<commit>."
        )
    if len(latest_tags) > 1:
        latest = get_latest_by_timestamp(opener, parts, image, label, latest_tags)
    elif len(latest_tags) == 1:
        latest = latest_tags[0]
    else:
        raise RuntimeError(
            f"The requested image {image} does not have any tags corresponding to the "
            f"expected format {base_tag}_<date>_<commit>."
        )
    return str(latest)
